/****************************************************************
 *
 * Proprietary Data of Adaptive Networks, Inc.
 *              All rights reserved
 *
 * Notice:
 *		This file contains trade secret data which
 *		is the property of Adaptive Networks, Inc.  This
 *		file may not be used for any purpose unless
 *		permission is granted in writing by an officer
 *		of Adaptive Networks, Inc.
 *
 *
 * Filename
 *		dataAPI.c
 *
 * Description: Handles DATA_HEADER structure (see MackLink.h)
 *					- allocate/free/Reallocate/Align data
 *					- initialize data header
 *					- read/write to the data header
 *				All functions comes in 4 flavors:
 *					- uint64_t data header
 *					- uint32_t data header
 *					- uint16_t data header
 *					- uint8_t  data header
 *					- uint1_t  data header
 *
 * Note: Many run-time support (RTS) functions use lock and unlock
 *	 	 functions to prevent reentrancy. However, DSP/BIOS SWI and
 *       HWI threads cannot call LCK_pend and LCK_post. As a result,
 *       RTS functions that call LCK_pend or LCK_post must not be
 *       called in the context of a SWI or HWI thread.
 *		 The following table lists some RTS functions that call
 *		 LCK_pend and LCK_post in certain versions of Code Composer
 *		 Studio: LCK_pend Acquire ownership of a resource lock
 *
 *			fprintf 	printf 		vfprintf 	sprintf
 *			vprintf 	vsprintf 	clock 		strftime
 *			minit 		malloc 		realloc 	free
 *			calloc 		rand 		srand 		getenv
 *
 * Revision History:
 *	Khaled Saab
 *	- Nov. 26th 2010 ported from the Cortex
 *	Mustapha Ait Yakoub
 *	- June 7th  2011 functions which refer to 64 bits were added
 *
 ****************************************************************/
#include <stdint.h>  		//uint32_t, ...
#include <stdlib.h>			//calloc
#include <string.h>			//memcpy

#include "../Shared/DSP1000.h"
#include "../Shared/dataAPI.h"

/**************************************************************************
* Function     : dataAdd08
* Author       : khaled saab
* Email        : saab@adaptiveNetworks.com
*
* Description  : Initialize the data header.
* Input        : data_p  : pointer to the data header
*				 length  : array size
*
* Output       : set the structure.
* Return       : SUCCESS/FAIL
* Revision     : Creation Date: Nov 24th 2009
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(dataAdd08,"AdaptiveCommonCode")
#endif
uint16_t dataAdd08( DATA_HEADER08 *data_p,
                    uint16_t maxLength) {

    //sanity check
	if(data_p == NULL) return FAIL;

	//reset all
    dataInitToDefault08(data_p);
	if(maxLength == 0) return SUCCESS;

    //allocate the memory and configure the block to default setting
    data_p->ptr = (uint8_t *)malloc(maxLength * sizeof(uint8_t));

    if(data_p->ptr == NULL) {
        return FAIL;
    }

    //return
    data_p->ptrSave   = data_p->ptr;  //copy the pointer for future use
    data_p->maxLength = maxLength;

    return SUCCESS;
}
/**************************************************************************
* Function     : dataRemove08
* Author       : khaled saab
* Email        : saab@adaptiveNetworks.com
*
* Description  : free the data header
*
* Input        : data_p : pointer to the data header
* Output       : set the structure.
* Return       : SUCCESS/FAIL
* Revision     : Creation Date: Nov 24th 2009
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(dataRemove08,"AdaptiveCommonCode")
#endif
uint16_t dataRemove08(DATA_HEADER08 *data_p) {

    //sanity check
	if(data_p          == NULL) return FAIL;

	if(data_p->ptrSave != NULL)	{
#ifdef __TI_COMPILER_VERSION__
#ifdef DEBUG_ALLOC_FREE
		if (data_p->ptrSave < LL2RAM_HEAP_BASE || data_p->ptrSave > LL2RAM_HEAP_END) {
			// Trying to free memory outside of heap boundary
			gDataFreePtrErrors++;
			return FAIL;
		} else {
			pFreeSize = (uint32_t *)(data_p->ptrSave) - 1;
			if ( *pFreeSize != (data_p->maxLength + 8) ) {
				gDataFreeSizeErrors++;
			}
#endif
#endif

			free(data_p->ptrSave);

#ifdef __TI_COMPILER_VERSION__
#ifdef DEBUG_ALLOC_FREE
			gMemAllocationHistory[gMemHistoryIndex].pData = data_p->ptrSave;
			gMemAllocationHistory[gMemHistoryIndex].allocationSize = NULL;
			gMemHistoryIndex++;
//			gMemAllocationHistory[gMemHistoryIndex].callingFunctionRecorded = FALSE;
		}
#endif
#endif
	}
    dataInitToDefault08(data_p);

	//return
	return SUCCESS;
}
/**************************************************************************
* Function     : dataCopy08
* Description  : copy "length" bytes from the dataSrc_p[dataSrcIndex]
*				 to dataDst_p[dataDstIndex]
* Input        : dataSrc_p		: pointer to the source data
*				 dataSrcIndex	: reading index
*				 dataDst_p		: pointer to the destination data
*				 dataDstIndex	: writting index
*				 length			: number of bytes to copy
* Input/Output :
* Return       : SUCCESS/FAIL
* History      :
*	Khaled Saab	Nov 06th 2013 creation date
*	Khaled Saab Mar 28th 2014 could not copy in place. Had to go through
*							  temp memory.
*	Khaled Saab Jul 02th 2015 set the length field of the copied data
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(dataCopy08,"AdaptiveCommonCode")
#endif
uint16_t dataCopy08(DATA_HEADER08 *dataSrc_p,
					uint16_t       dataSrcIndex,
					DATA_HEADER08 *dataDst_p,
					uint16_t       dataDstIndex,
					uint16_t       length) {

	DATA_HEADER08 dataTmp;

	//sanity check
	if(length    == 0) 										return SUCCESS;
	if(dataSrc_p == NULL)									return FAIL;
	if(dataDst_p == NULL)									return FAIL;
	if(dataSrcIndex + length > dataMaxLength08(dataSrc_p)) 	return FAIL;
	if(dataDstIndex + length > dataMaxLength08(dataDst_p)) 	return FAIL;

	//copy the data to tmp location
	if(dataAdd08(&dataTmp, length) != SUCCESS)
		return FAIL;

	memcpy(	dataPtr08(&dataTmp, 0), 				//dst
			dataPtr08(dataSrc_p, dataSrcIndex), 	//src
			length * sizeof(uint8_t));				//length

	//copy the data to final location
	memcpy(	dataPtr08(dataDst_p, dataDstIndex),		//dst
			dataPtr08(&dataTmp, 0),					//src
			length * sizeof(uint8_t));				//length
	dataSetLength08(dataDst_p, dataDstIndex + length);

	//free the memory
	dataRemove08(&dataTmp);

	//return
	return SUCCESS;
}
/**************************************************************************
* Function     : dataInitToDefault08
* Description  : init the struct to default values
*
* Input        : data_p : pointer to the data header
* Output       :
* Return       :
* Revision     :
*       khaled Saab     Dec 1th 2010 Creation Date
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(dataInitToDefault08,"AdaptiveCommonCode")
#endif
void dataInitToDefault08(DATA_HEADER08 *data_p) {

    //sanity check
	if(data_p == NULL) return;

    data_p->ptr       = NULL;
    data_p->ptrSave   = NULL;
    data_p->format    = WORD_SIZE_08_BITS;
    data_p->putIndex  = 0;
    data_p->getIndex  = 0;
    data_p->RWIndex   = 0;
    data_p->length    = 0;
    data_p->maxLength = 0;
}
#ifdef _DEBUG
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(dataWrite08,"AdaptiveCommonCode")
#endif
void dataWrite08(DATA_HEADER08 *data_p, uint8_t val, int16_t index) {

	//sanity check
	if(data_p == NULL) 				while(1) { };
	if(index >= data_p->maxLength)	while(1) { };

	data_p->ptr[index] = val;
}
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(dataRead08,"AdaptiveCommonCode")
#endif
uint8_t dataRead08(DATA_HEADER08 *data_p, int16_t index) {

	//sanity check
	if(data_p == NULL) 				while(1) { };
	if(index >= data_p->maxLength)	while(1) { };

	return (data_p->ptr[index]);
}
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(dataPut08,"AdaptiveCommonCode")
#endif
void dataPut08(DATA_HEADER08 *data_p, uint8_t val) {
    dataWrite08(data_p, val, data_p->putIndex++);
}

#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(dataGet08,"AdaptiveCommonCode")
#endif
uint8_t dataGet08(DATA_HEADER08 *data_p) {
    return (dataRead08(data_p, data_p->getIndex++));
}
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(dataSetLength08,"AdaptiveCommonCode")
#endif
void dataSetLength08(DATA_HEADER08 *data_p, uint16_t val){

	if(val <= (data_p)->maxLength) {
		(data_p)->length = val;
	}
	else {
		(data_p)->length = (data_p)->maxLength;
	}
}
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(dataSetGetIndex08,"AdaptiveCommonCode")
#endif
void dataSetGetIndex08(DATA_HEADER08 *data_p, uint16_t val){
	if(val <= (data_p)->length) {
		(data_p)->getIndex = val;
	}
	else {
		(data_p)->getIndex = (data_p)->length;
	}
}
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(dataSetPutIndex08,"AdaptiveCommonCode")
#endif
void dataSetPutIndex08(DATA_HEADER08 *data_p, uint16_t val){
	if(val <= (data_p)->length) {
		(data_p)->putIndex = val;
	}
	else {
		(data_p)->putIndex = (data_p)->length;
	}
}
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(dataSetRWIndex08,"AdaptiveCommonCode")
#endif
void dataSetRWIndex08(DATA_HEADER08 *data_p, uint16_t val){
	if(val <= (data_p)->length) {
		(data_p)->RWIndex  = val;
	}
	else {
		(data_p)->RWIndex  = (data_p)->length;
	}
}
#endif
