/****************************************************************
  *
  * Proprietary Data of Adaptive Networks, Inc.
  *              All rights reserved
  *
  * Notice:
  *        This file contains trade secret data which
  *        is the property of Adaptive Networks, Inc.  This
  *        file may not be used for any purpose unless
  *        permission is granted in writing by an officer
  *        of Adaptive Networks, Inc.
  *
  *
  * Filename
  *        packetAPI.c
  *
  * Description: this file contains all functions that are used to
  *				 access a packet (see MacLink.h):
  *					- allocate/free/Reallocate/Align packets
  *					- initialize a packet
  *					- read/write to the packet
  *
  * Note: Many run-time support (RTS) functions use lock and unlock
  *		 functions to prevent reentrancy. However, DSP/BIOS SWI and
  *      HWI threads cannot call LCK_pend and LCK_post. As a result,
  *      RTS functions that call LCK_pend or LCK_post must not be
  *      called in the context of a SWI or HWI thread.
  *		 The following table lists some RTS functions that call
  *		 LCK_pend and LCK_post in certain versions of Code Composer
  *		 Studio: LCK_pend Acquire ownership of a resource lock
  *
  *			fprintf 	printf 		vfprintf 	sprintf
  *			vprintf 	vsprintf 	clock 		strftime
  *			minit 		malloc 		realloc 	free
  *			calloc 		rand 		srand 		getenv
  *
  * Revision History:
  *     khaled Saab     Oct 16th 2010   creation date
  ****************************************************************/
#include <stdint.h>				//declaration of uint8_t int16_t...
#include <stdio.h>
#include <string.h>				//memcpy

#include <stdbool.h>			//bool
#include <driverLib/interrupt.h>//interrupt enable/disable

#include "../Shared/DSP1000.h"
#include "../Shared/MacLink.h"
#include "../Shared/dataAPI.h"
#include "../Shared/packetAPI.h"

//-----------------------------------------------------------------
//Global variables
//-----------------------------------------------------------------
PACKET_HEADER gPacketHeader[PACKET_MAX_NUM];
uint16_t      gPacketAvailable = PACKET_MAX_NUM;
uint16_t      gPacketIndex     = 0;		//last packet used

#ifdef CORTEX_APP
extern OS_MUT 	gMutexFlag;				//Mutex used to lock the packet access
#endif

/**************************************************************************
* Function     : packetGet
*
* Description  : get the next word
*				 This function will increment the packet getIndex by 1.
*
* Input        : PACKET_HEADER   : pointer to the serial data
* Output       : word            : the word
* Return       : if success, return the word, else return 0
* History      :
*       khaled Saab     Oct 19th 2010 creation data
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(packetGet,"AdaptiveCommonCode")
#endif
uint32_t packetGet(PACKET_HEADER *packet_p) {
    return dataGet_(&(packet_p->data));
}
/**************************************************************************
* Function     : packetPut
*
* Description  : put the input word.
*				 This function will increment the packet putIndex by 1.
*
* Input        : word             : the word
* Input/Output : PACKET_HEADER    : pointer to packet
* Return       : SUCCESS/FAIL
* History      :
*       khaled Saab     Oct 19th 2010 creation data
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(packetPut,"AdaptiveCommonCode")
#endif
uint16_t packetPut(PACKET_HEADER *packet_p, uint32_t word) {

	//sanity check
	if(packetGetPtr(packet_p) == NULL)
		return FAIL;

    dataPut_(&(packet_p->data), word);
    return SUCCESS;                          //return
}
/**************************************************************************
* Function     : packetRead
*
* Description  : read a word from a packet.
* Input        : PACKET_HEADER    : pointer to packet
*                index            : word index
* Output       : word             : the word
* Return       : if success, return the word, else return 0
* History      :
*       khaled Saab     Oct 19th 2010 creation data
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(packetRead,"AdaptiveCommonCode")
#endif
uint32_t packetRead(PACKET_HEADER *packet_p, int16_t index) {
    return dataRead_(&(packet_p->data), index);
}
/**************************************************************************
* Function     : packetWrite
*
* Description  : Write the input word to a packet.
* Input        : word             : the word to write
*				 index            : index of word.
* Input/Output : PACKET_HEADER    : pointer to packet
* Return       : SUCCESS/FAIL
* History      :
*       khaled Saab     Oct 19th 2010 creation data
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(packetWrite,"AdaptiveCommonCode")
#endif
uint16_t packetWrite(PACKET_HEADER *packet_p, uint32_t word, int16_t index) {

	//sanity check
	if(packetGetPtr(packet_p) == NULL)
		return FAIL;

    dataWrite_(&(packet_p->data), word, index);
    return SUCCESS;
}
/**************************************************************************
* Function     : packetAvailableCount
*
* Description  : return the total noumber of packets available.
*				 This function could be used in conjunction with
*				 gPacketAvailable to figure out how many packets are available
*				 In general case, use gPacketAvailable, but if a partial reset
*				 is needed, this function could be used.
* Input        :
* Input/Output :
* Return       : number of packets available.
* History      :
*       khaled Saab     Feb 21th 2013 creation data
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(packetAvailableCount,"AdaptiveCommonCode")
#endif
uint16_t packetAvailableCount(void) {
	uint16_t i;
	uint16_t count = 0;
	VAR_CRITICAL;		//save and restore HWI info

	//access to shared resources needs to be performed atomically
	ENTER_CRITICAL();

	for(i = 0; i < PACKET_MAX_NUM; i++)
		if(packetState(&(gPacketHeader[i])) == PACKET_DEFAULT_STATE)
			count ++;

	//return
	EXIT_CRITICAL();
    return count;
}
/**************************************************************************
* Function     : packetAcquire()
*
* Description  : obtain a packet from the packet array
* Input        : dataLength     : length of the data in the packet
* Output  	   : pointer to the packet.
*                if the function failes to get a packet with the
*                specified data length, NULL is returned.
* Revision History:
*	Khaled Saab	Oct 16th 2010   Migrated from PktMgrCreatePacket()
*	Khaled Saab	Jun 27th 2012	Made the critical section atomic
*	Khaled Saab Aug  5th 2013	Change the search loop to start from
*								where we left off.
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(packetAcquire,"AdaptiveCommonCode")
#endif
PACKET_HEADER *packetAcquire(uint16_t dataLength) {

	 int16_t     i;
	 int16_t     packetIndex;
	uint16_t     packetFound;
	DATA_HEADER_ data;
	VAR_CRITICAL;		//save and restore HWI info

	//check if there is more packet
	//scan all packets for an empyt one. if we reach the
	//end, and no packet is empty, return FAIL
	if(gPacketAvailable == 0) {
		gPacketAvailable = packetAvailableCount();	//extra sanity check
		return NULL;
	}

	//access to shared resources needs to be performed atomically
	//disable all interrupts before entering the "for" loop
	ENTER_CRITICAL();

	//get a valid packet index
	for(i = 0, packetFound = FALSE; i < PACKET_MAX_NUM; i++) {
		packetIndex  = i + gPacketIndex + 1;
		packetIndex %= PACKET_MAX_NUM;

		if(gPacketHeader[packetIndex].status.state == PACKET_DEFAULT_STATE) {

			//set the status field so we don't reuse the same header
			gPacketHeader[packetIndex].status.state = PACKET_IN_USE_STATE;

			packetFound = TRUE;
			break;
		}//if we found a valid packet
	}//search for a valid packet

	if(packetFound == FALSE) {
		goto end_failCondition;
	}

	//allocate the memory for the data
	if(FAIL == dataAdd_(&data, dataLength)) {
		gPacketHeader[packetIndex].status.state = PACKET_DEFAULT_STATE;
		goto end_failCondition;
	}
	gPacketHeader[packetIndex].data = data;

	gPacketIndex = packetIndex;
	if(gPacketAvailable > 0)
		gPacketAvailable --;

#ifdef DEBUG_ALLOC_FREE
	//Last sanity check
	//This is for debug purposes and could be removed once the code
	//is stable
	if(packetCollision(&gPacketHeader[pktIdx])) {
		logError(PACKET_LENGTH_FIELD_ERROR);

		packetRemove(&gPacketHeader[packetIndex]);
		gPacketAvailable = packetAvailableCount();

		goto end_failCondition;
	}
#endif

	//return
	EXIT_CRITICAL();
	return (&(gPacketHeader[packetIndex]));

end_failCondition:
	EXIT_CRITICAL();
	return NULL;
}
/**************************************************************************
* Function     : packetCopy
*
* Description  : copy the packet pointed to by src_p to dst_p.
* Input        : packetDst_pp : packet destination pointer
*                packetSrc_p  : packet source pointer
* Output       :
* Return       : SUCCESS/FAIL
* Note         : the packet state is not copied
*				 the packet queue is not copied
*				 memory aligned is not maintained
* Note         : if data is cached, memcpy work with the cashed data i.e.
*				 copies datas to/from the cashe to their destinations one
*				 cache line at a time.
*				 if the data is not cached, memcopy copy the data to/from
*				 the initial destination/source address.
* Revision     :
* Khaled Saab Oct 29th 2012 Creation Date
* Khaled Saab Feb 22th 2016	Removed length sanity check since it was
*							blocking large packets.
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(packetCopy,"AdaptiveCommonCode")
#endif
uint16_t packetCopy(PACKET_HEADER **packetDst_pp,
					PACKET_HEADER *packetSrc_p){

	PACKET_HEADER *packetDst_p;
	uint8_t       *ptrSave;
	VAR_CRITICAL;		//save and restore HWI info

	//access to shared resources needs to be performed atomically
	ENTER_CRITICAL();

	//sanity check
	//Note: we are not using packetValid( ) because, in the DSP, the
	//		src packet could be from the ICC buffer.
	if(packetSrc_p  == NULL) 				goto end_failCondition;
	if(packetDst_pp == NULL) 				goto end_failCondition;
	if(packetLength(packetSrc_p) > packetMaxLength(packetSrc_p))
											goto end_failCondition;
	//allocate memory
	packetDst_p = packetAcquire(packetMaxLength(packetSrc_p));
	if(packetDst_p == NULL) 				goto end_failCondition;

	//save the dst packet pointer (memory allocation)
	ptrSave = packetGetPtr(packetDst_p);

	//copy the header
	//restore the dst packet pointer
	//don't copy the state or the queue
	memcpy(packetDst_p, packetSrc_p , sizeof(PACKET_HEADER));

	packetSetPtr(    packetDst_p, ptrSave);
	packetSetPtrSave(packetDst_p, ptrSave);
	packetState(     packetDst_p) = PACKET_IN_USE_STATE;
	packetQueue(     packetDst_p) = NULL;	//do not belong to any queue
	packetNext(      packetDst_p) = NULL;	//do not belong to any queue

	//copy the payload
	if(	(packetLength(packetSrc_p) != 0   ) &&
		(packetGetPtr(packetSrc_p) != NULL) &&
		(packetGetPtr(packetDst_p) != NULL) )
		memcpy(	packetGetPtr(packetDst_p),
				packetGetPtr(packetSrc_p),
				sizeof(uint8_t) * packetLength(packetSrc_p));

	//return
	EXIT_CRITICAL();
	*packetDst_pp = packetDst_p;
	return SUCCESS;

end_failCondition:
	//return
	EXIT_CRITICAL();
	*packetDst_pp = NULL;
    return FAIL;
}
/**************************************************************************
* Function     : packetAdd
*
* Description  : allocate memory for  a packet and return it for use.
* Input        : packet_p	: pointer to the packet header to be add
*                maxLength  : array size
* Output       : set the packet header structure to default values
* Return       : SUCCESS/FAIL
*
*
* Warning      : This function is provided for convenience and testing.
*				User should be carefull using this function.
*				This function is NOT part of the standard packet API and
*				thus its NOT managed by it.
*				It was added to give the user the capability to create
*				his own packet.
*
* Revision     :
*       khaled Saab Oct16th 2010 : Creation Date
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(packetAdd,"AdaptiveCommonCode")
#endif
uint16_t packetAdd(PACKET_HEADER *packet_p, uint16_t maxLength) {

	uint16_t status;

    //sanity check
    if(packet_p == NULL)
        return FAIL;

	//initialization
	packetInitToDefault(packet_p);

    //allocate the memory
    status = dataAdd_(&(packet_p->data),maxLength);
	if(status == FAIL)
		return FAIL;

    gPacketAvailable--;
	return SUCCESS;
}
/**************************************************************************
* Function     : packetRemove
*
* Description  : if the packet dosn't belong to a queue
*					Free the packet
*				else
*					clear the packet TTL (so it can be removed by the TTL
*					manager).
* Input        : packet_p	: pointer to the packet header to be freed
* Output       : clear the packet header structure.
* Return       : SUCCESS/FAIL. Return FAIL if the packet belongs to a queue.
*
* Revision     :
*	Khaled Saab Oct16th 2010 	Creation Date
*	Khaled Saab	July 5th 2012	removed the packet only if it dosn't belong
*								to a queue
*	Khaled Saab	Nov 16th 2012	made the function atomic
*	Khaled Saab	Nov 29th 2012	added a check for packetValid( )
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(packetRemove,"AdaptiveCommonCode")
#endif
uint16_t packetRemove(PACKET_HEADER *packet_p) {

	uint16_t status;
	VAR_CRITICAL;		//save and restore HWI info

	//sanity check
	if(!packetValid(packet_p)) 	return FAIL;

	//access to shared resources needs to be performed atomically
	//disable all interrupts before entering the "for" loop
	ENTER_CRITICAL();

#ifdef DEBUG_ALLOC_FREE
	if(packetCollision(packet_p)) {
		logError(PACKET_LENGTH_FIELD_ERROR);

		gPacketAvailable = packetAvailableCount();

		EXIT_CRITICAL();
		return FAIL;
	}
#endif

	//free the packet payload
	packetDataRemove(packet_p);				//free the memory

	//if the packet belongs to a queue
	//	clear the TTL and let the garbage collector handle it
	//else
	//	its safe to remove the packet
	if(packetQueue(packet_p) == NULL) {		//packet does not belong to a queue
		packetInitToDefault(packet_p);		//reset all packet variables
		if(gPacketAvailable < PACKET_MAX_NUM)
			gPacketAvailable++;
		status = SUCCESS;
	}
	else {									//packet belongs to a queue
		status = FAIL;
	}

	//restore interrupts
	EXIT_CRITICAL();
	return status;
}
/**************************************************************************
* Function     : packetInit()
*
* Description  : Initialize all packets gPacketHeader[PACKET_SIZE] to default
*                values, but do not allocate memory for the data it self.
*                The memory allocation for the data will be done in packetGet( )
* Input        :
* Output  	   :
* Revision History:
*		khaled Saab     Oct 16th 2010   initial version
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(packetInit,"AdaptiveCommonCode")
#endif
void packetInit(void) {

	int16_t i;

	//initialize all packets to default values
	for(i = 0; i < PACKET_MAX_NUM; i++) {
		packetInitToDefault(&(gPacketHeader[i]));
	}

	gPacketAvailable = PACKET_MAX_NUM;
	gPacketIndex     = 0;
}
/**************************************************************************
* Function     : packetInitToDefault()
*
* Description  : Initialize a packets to default values
*                But do not allocate memory for the data it self.
*                The memory allocation for the data will be done in packetGet( )
*
* Input        : packet_p   : pointer to the packet to init
* Output  	   :
*
* Revision History:
*		khaled Saab     Oct 16th 2010   initial version
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(packetInitToDefault,"AdaptiveCommonCode")
#endif
void packetInitToDefault(PACKET_HEADER *packet_p) {

    //before changing anything, make sure that the packet is
	//not accessible
	packet_p->status.state = PACKET_IN_USE_STATE;

    //do not allocate the packet payload yet
    dataInitToDefault_(&(packet_p->data));

    //Interface (Input/Output)
    packet_p->endPoint.srcEndPoint  = 0;
    packet_p->endPoint.dstEndPoint  = 0;
    packet_p->endPoint.addrEndPoint = 0;
    packet_p->endPoint.command      = 0;

    packet_p->status.next           = NULL;
	packet_p->status.queue          = NULL;		//doesn't belong to any queue yet
    packet_p->status.state          = PACKET_DEFAULT_STATE;
}
/**************************************************************************
* Function     : packetValid()
*
* Description  : check if its a valid packet
*
* Input        : packet_p : pointer to the packet
* Output  	   :
* Return       : TRUE/FALSE: TRUE if its a valid packet and false otherwise.
* Revision History:
* Khaled Saab	Nov 29th 2012	initial version.
* Khaled Saab	Aug 22th 2013	added compile flags to disable the scan
*								in release mode.
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(packetValid,"AdaptiveCommonCode")
#endif
uint16_t packetValid(PACKET_HEADER *packet_p) {

	#ifdef _DEBUG
	uint16_t i;
	#endif

	//initial quick test to speed up processing.
	if(packet_p == NULL) return FALSE;

	#ifdef _DEBUG
	//scan for a matching packet
	for(i = 0; i <  PACKET_MAX_NUM; i++) {
		if(packet_p == (&(gPacketHeader[i])) )
			return TRUE;
	}
	return FALSE;

	#else
	return TRUE;
	#endif
}
