/****************************************************************
  *
  * Proprietary Data of Adaptive Networks, Inc.
  *              All rights reserved
  *
  * Notice:
  *        This file contains trade secret data which
  *        is the property of Adaptive Networks, Inc.  This
  *        file may not be used for any purpose unless
  *        permission is granted in writing by an officer
  *        of Adaptive Networks, Inc.
  *
  *
  * Filename
  *        queueAPI.c
  *
  * Description: contains queue management code
  *
  * Revision History:
  *     khaled Saab     Oct 16th 2010   moved from queue_driver.c
  ****************************************************************/
#include <stdint.h>				//declaration of uint8_t int16_t...

#ifdef __TI_COMPILER_VERSION__
#include <std.h>				//boolean, uint...
#include <c6x.h>
#include <hwi.h>				//interrupt enable/disable
#endif
#ifdef CORTEX_APP
#include <stdbool.h>		 	//bool
#include <inc/hw_types.h>		//boolean, uint...
#include <driverLib/interrupt.h>//interrupt enable/disable
#include "../Cortex/main/main.h"//stats
#endif

#include "../Shared/DSP1000.h"
#include "../Shared/MacLink.h"
#include "../Shared/dataAPI.h"
#include "../Shared/packetAPI.h"
#include "../Shared/queueAPI.h"


//-----------------------------------------------------------------
//Global variables
//-----------------------------------------------------------------
//queues that runs on the Cortex core
QUEUE_HEADER   gQueueEthernetTx  = QUEUE_HEADER_DEFAULT;   //send to Ethernet
QUEUE_HEADER   gQueueEthernetRx  = QUEUE_HEADER_DEFAULT;   //receive from Ethernet

#ifdef CORTEX_APP
extern OS_MUT  gMutexFlag;			//Mutex used to lock the queue access
#endif

/**************************************************************************
* Function     : queueInit()
*
* Description  : Initialize all queues
*                For each ENDPOINT, allocated and initialize the queue headers.
*                Typically two queue per endpoint are allowed: one for Tx and
*                another for Rx.
*
* Input        :
* Output  	   :
* Note         :
*
* Revision History:
*		khaled Saab     Oct 16th 2010   initial version
*       khaled Saab     Nov 18th 2010   added Ethernet queue
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(queueInit,"AdaptiveCommonCode")
#endif
void queueInit(void) {
    //initialize the Ethernet queue(s)
    if(ENDPOINT_ETHERNET_TX_ENABLE)			queueInitToDefault(&gQueueEthernetTx);
    if(ENDPOINT_ETHERNET_RX_ENABLE)			queueInitToDefault(&gQueueEthernetRx);
}
/**************************************************************************
* Function     : queueInitToDefault()
*
* Description  : Initialize a single queue header to default value
* Input        : *queueHeader_p     : pointer to the queue header
*                maxLength          : maximal number of packets in this queue
*                burstAllowed       : if TRUE burst mode is allowed
* Output  	   :
* Note         :
*
* Revision History:
*	Khaled Saab	Oct 16th 2010   initial version
*	Khaled Saab	May  5th 2013	imbeded the max length and burst size in
*								the function for more flexibility (no
*								need to provide them as input arguments).
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(queueInitToDefault,"AdaptiveCommonCode")
#endif
void queueInitToDefault(QUEUE_HEADER *queueHeader_p) {

	uint32_t maxLength;
	uint32_t burstSize;
	HWI_MasterVariable;

	//sanity check
	if(!queueValid(queueHeader_p))	return;

	//initialization
	maxLength = 0;
	burstSize = 0;


	if(queueHeader_p == &gQueueEthernetTx) {
		maxLength = ENDPOINT_ETHERNET_TX_MAX_LENGTH;
		burstSize = ENDPOINT_ETHERNET_BURST_SIZE;
	}
	else if(queueHeader_p == &gQueueEthernetRx) {
		maxLength = ENDPOINT_ETHERNET_RX_MAX_LENGTH;
		burstSize = ENDPOINT_ETHERNET_BURST_SIZE;
	}

	//make this section of the code atomic
	HWI_MasterDisable();
	queueNext(queueHeader_p)        = NULL;        //no packet in the queue
    queueActivePacket(queueHeader_p)= NULL;        //no packets is beeing processed

    queueLength(queueHeader_p)      = 0;           //no packets are in the queue yet
    queueMaxLength(queueHeader_p)   = maxLength;   //maximal number of packets in this queue
    queueBurstSize(queueHeader_p)   = burstSize;   //enable/disable burst mode
	HWI_MasterEnable();
}
/**************************************************************************
* Function     : queueValid()
*
* Description  : return TRUE is the queue address is pointing to a valid queue
*				 and FALSE otherwise
* Input        : *queueHeader_p     : pointer to the queue header
* Output  	   :
* return       : TRUE/FASLE	: TRUE if the queue is valid
*
* Revision History:
*	khaled Saab     May 28th 2012    creation date
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(queueValid,"AdaptiveCommonCode")
#endif
uint16_t queueValid(QUEUE_HEADER *queueHeader_p) {

	//to speed things up
	if(queueHeader_p == NULL)
		return FALSE;

    return((queueHeader_p == &gQueueEthernetTx)  ||
           (queueHeader_p == &gQueueEthernetRx));

	#ifdef BOOTLOADER
	return TRUE;
#endif
}
/**************************************************************************
* Function     : queueSanityCheck()
*
* Description  : perform a queue sanity check
*					check queue length
*					check if all packets are valid
*					check if all packets belongs to the queue
*
* Input        : *queueHeader_p     : pointer to the queue header
* Output  	   :
* Return       : SUCCESS/FAIL. Return SUCCESS if the queue passed the
*				 sanity check and FAIL otherwise.
* Note         :
*
* Revision History:
*	khaled Saab	March 26th 2013   initial version
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(queueSanityCheck,"AdaptiveCommonCode")
#endif
uint16_t queueSanityCheck(QUEUE_HEADER *queueHeader_p) {

	uint16_t       i;
	PACKET_HEADER *packet_p;
	uint16_t       nbPacketsInQueue;
	VAR_CRITICAL;		//save and restore HWI info

	//sanity check
	if(!queueValid(queueHeader_p))	return FAIL;

	//make this section of the code atomic
	ENTER_CRITICAL();

	//clear the queue
	packet_p         = queueNext(queueHeader_p);
	nbPacketsInQueue = queueLength(queueHeader_p);
	for(i = 0; i < nbPacketsInQueue; i++) {

		if(!packetValid(packet_p)) 					goto end_failCondition;
		if( packetQueue(packet_p) != queueHeader_p) goto end_failCondition;

		packet_p = packetNext(packet_p);
	}//for all packets in the queue

	//end of atomic section
	EXIT_CRITICAL();

	//return
	if(packet_p != NULL)	return FAIL;
    else					return SUCCESS;

end_failCondition:
	EXIT_CRITICAL();
	return FAIL;
}
/**************************************************************************
* Function     : queueReset()
*
* Description  : reset the queue header and clear TTL of all packets that
*				 belongs to this queue.
* Input        : *queueHeader_p     : pointer to the queue header
* Output  	   :
* Return       : SUCCESS/FAIL
* Note         :
*
* Revision History:
*	Khaled Saab	Feb 2th 2011 	initial version
*	Khaled Saab	Jun 27th 2012	Since we don't want this function to be atomic
*								(its running from a low level task) added
*								PACKET_LOCK_STATE flag to simulate atomicity.
*	Khaled Saab Nov 26th 2012	made the call to queueInitToDefault( ) function
*								atomic: 1) cant make the entire function atomic
*								because it calling queueRemove( )
*	Khaled Saab Feb 13th 2013	made the function atomic.
*								removed PACKET_LOCK_STATE.
*	Khaled Saab May 17th 2013	simplified the function and added an extra
*								loop to scan the packets for broken links.
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(queueReset,"AdaptiveCommonCode")
#endif
uint16_t queueReset(QUEUE_HEADER *queueHeader_p){

	uint16_t       i;
	QUEUE_HEADER  *packetQueue_p;
	PACKET_HEADER *packet_p;

	VAR_CRITICAL;		//save and restore HWI info

	//sanity check
	if(!queueValid(queueHeader_p))
		return FAIL;

	//make this section of the code atomic
	ENTER_CRITICAL();

	//reset the queue
	queueInitToDefault(queueHeader_p);

	//scan all packets for broken links
	for(i = 0; i < PACKET_MAX_NUM; i++) {
		packet_p      = &(gPacketHeader[i]);
		packetQueue_p = packetQueue(packet_p);

		if(packetState(packet_p) == PACKET_DEFAULT_STATE)
			continue;

		//if the packet belongs to this queue OR
		//if the packet do not belong to  any of the valid queues
		//	schedule it for removal
		//Note: packet which has a NULL queue should not be reset
		//		since they could be valid packets that has not been
		//		assigned a queue yet.
		if(	(packetQueue_p == queueHeader_p) ||
			((packetQueue_p != NULL) && (!queueValid(packetQueue_p)))) {

			packetQueue(packet_p) = NULL;	//do not belong to the queue anymore
		}
	}

	//return
	EXIT_CRITICAL();
    return SUCCESS;
}
/**************************************************************************
* Function     : queueClearTTL()
*
* Description  : Set the timeout flag for all packets in the queue
* Input        : *queueHeader_p     : pointer to the queue header
* Output  	   :
* Return       : SUCCESS/FAIL
* Note         : QueueRemoveExpiredTTL will free the packet and its payload.
* Note         : since this function could be called from an ISR (Cortex)
*				 we can't call Mutex wait( ) and Mutex release( ). As such
*				 this function can't be atomic.
* Revision History:
*	Khaled Saab	March 19th 2011	initial version
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(queueClearTTL,"AdaptiveCommonCode")
#endif
uint16_t queueClearTTL(QUEUE_HEADER *queueHeader_p){

	uint16_t       i;
	uint16_t       nbPacketsInQueue;
 	PACKET_HEADER *currPacket_p;
 	PACKET_HEADER *nextPacket_p;	//just in case the current packet
									//was removed just after its TTL
									//was set to "0"

//	VAR_CRITICAL;		//save and restore HWI info

	//sanity check
	if(!queueValid(queueHeader_p))
		return FAIL;

	//access to shared resources needs to be performed atomically
//	ENTER_CRITICAL();	//this function is called from ISR (Cortex)

	//clear the queue
	currPacket_p = queueNext(queueHeader_p);
	nextPacket_p = queueNext(queueHeader_p);

	nbPacketsInQueue = queueLength(queueHeader_p);
	for(i = 0; i < nbPacketsInQueue; i++) {

		currPacket_p = nextPacket_p;
		nextPacket_p = packetNext(currPacket_p);

		//sanity checks and additional stop condition
		//all packets in the queue should be valid packets
		//all packets in the queue should be pointing back to the queue header
		if( !packetValid(currPacket_p)) 						break;
		if(  packetQueue(currPacket_p) != queueHeader_p)		break;

	}//for all packets in the queue

	//return
//	EXIT_CRITICAL();	//this function is called from ISR (Cortex)
	return SUCCESS;
}
/**************************************************************************
* Function     : queueGet()
*
* Description  : get the next packet from the queue.
*                The queue is a sorted list with highest priority at
*                the top.
*						packetStatus &   setFlag == setFlag &&
*						packetStatus & clearFlag == 0
*
* Input        : *queueHeader_p     : pointer to the queue header
*                 setFlag           : status flag that should match the
*									  packet.
*                                     its a logical or between the packet
*									  state flags (see below)
*
*				 clearFlag	 		: status flag that should not be set
*									  in order to retreive the packet.
*									  Set to NULL to bypass.
*									  Same as the set flags, its a logic
*									  or between the packet state flags
*									  (see below).
*
*                      PACKET_DEFAULT_STATE : return all
*                      PACKET_IN_USE_STATE  : return active packets
*                      PACKET_READY_STATE   : return the packets that are
*											  ready for processing
*                      PACKET_ACTIVE_STATE  : return the packets that are
*                                             currently being processed
*                      PACKET_DONE_STATE    : return the packet that has
*											  been processed (History)
*                      PACKET_TTL_EXPIRED_STATE : return the packets with
*                                             expired TTL
* Output  	   :
* return       : *packetHeader_p    : pointer to the packet with matching
*									  input flag
* Note         : this function will not return packets with expired TTL,
*				 unless they are active packets.
* Revision History:
* Khaled Saab 	Nov 2th 2010    creation date
* Khaled Saab 	Apr 17th 2014 	added input argument: the clear flags (needs
*								to be cleared for the packet to be retreived).
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(queueGet,"AdaptiveCommonCode")
#endif
PACKET_HEADER *queueGet(QUEUE_HEADER *queueHeader_p,
						uint16_t setFlag,
						uint16_t clearFlag) {

    uint16_t i;
	uint16_t found;
	uint16_t nbPacketsInQueue;		//tmp place holder for queueLength( )
    PACKET_HEADER *packet_p;

	//sanity check
	if(!queueValid(queueHeader_p)) return NULL;

	found            = FALSE;
	packet_p         = queueNext(queueHeader_p);
	nbPacketsInQueue = queueLength(queueHeader_p);
	for(i = 0; i < nbPacketsInQueue; i++) {

		//sanity check
		//check if we are at the end of the queue
		//	the packet should be a valid packet
		//	the packet should point back to the queue header
		if(!packetValid(packet_p))						break;
		if( packetQueue(packet_p) != queueHeader_p) 	break;

		//if the flags match
		if( ((packetState(packet_p) &   setFlag) == setFlag) &&
			((packetState(packet_p) & clearFlag) == 0)) {

			found = TRUE;
			break;
		}//if flags match

		packet_p = packetNext(packet_p);
	}//for loop

	if(found)	return (packet_p);
	else		return NULL;
}
/**************************************************************************
* Function     : queueAdd()
*
* Description  : add a packet to a queue
*                the queue is a sorted list with highest priority at the
*                top. If more than 1 packet has the same priority, the
*                oldest will be at the top.
*
*                   search for a position to insert the packet into
*                   if there is no items in the queue
*                       insert at the top
*                   if the item priority is higher than than the one on the top
*                       insert at top of list
*                   we have exhausted the corner cases
*                   search for position to insert
*
* Input        : *queueHeader_p     : pointer to the queue header
*                *packetHeader_p    : pointer to the packet to insert
* Output  	   : SUCCESS/FAIL
*
* Revision History:
*       Warren Postmat  Feb 11th 2010   initial version (Queue_Driver.c)
*		khaled Saab     Oct 26th 2010   re-write in order to integrate with
*                                       packetAPI, dataAPI, and the task
*                                       manager.
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(queueAdd,"AdaptiveCommonCode")
#endif
uint16_t queueAdd(QUEUE_HEADER *queueHeader_p, PACKET_HEADER *packet_p){

	uint16_t i;
	uint16_t nbPacketsInQueue;		//tmp place holder for queueLength
    PACKET_HEADER *currPacket_p;
    PACKET_HEADER *nextPacket_p;

	VAR_CRITICAL;		//save and restore HWI info

	//access to shared resources needs to be performed atomically
	ENTER_CRITICAL();

    //sanity check
	//make sure the packet is valid, it does not already belongs to a queue
    //make sure we have enough space in the queue
	if(! packetValid(packet_p))					        			goto end_failCondition;
	if(!(packetState(packet_p) & PACKET_IN_USE_STATE))				goto end_failCondition;

	if( queueValid(packetQueue(packet_p)))							goto end_failCondition;
	if(!queueValid (queueHeader_p))									goto end_failCondition;

	if( queueLength(queueHeader_p) >= queueMaxLength(queueHeader_p)){
		if(queueSanityCheck(queueHeader_p) == FAIL)					goto end_breakDownCondition;
		else														goto end_failCondition;
	}

	//if there is no items in the queue
	//  insert at the top
	currPacket_p = queueNext(queueHeader_p);
	if(currPacket_p == NULL) {

		packetNext(packet_p)       = NULL;
		packetQueue(packet_p)      = queueHeader_p;	//packet belongs to this queue
		queueNext(queueHeader_p)   = packet_p;
		queueLength(queueHeader_p) = 1;
	}

	//we have exhausted the corner cases (there is at least one packet in the queue)
	//search for position to insert
	else {

		currPacket_p     = queueNext(queueHeader_p);//reset to top
		nextPacket_p     = queueNext(queueHeader_p);//reset to top
		nbPacketsInQueue = queueLength(queueHeader_p);
	    for(i = 0; i < nbPacketsInQueue; i++) {

			currPacket_p = nextPacket_p;
			nextPacket_p = packetNext(currPacket_p);

//			if(!packetValid(currPacket_p)) 					//redundant: should be valid
//				goto end_failCondition;
			if(!packetValid(nextPacket_p)) 					//end of the queue
				break;
			if(packetQueue(nextPacket_p) != queueHeader_p)	//sanity check
				goto end_breakDownCondition;
		}//search for packet

		packetNext (packet_p)     = nextPacket_p;
		packetQueue(packet_p)     = queueHeader_p;	//packet belongs to this queue
		packetNext (currPacket_p) = packet_p;
		queueLength(queueHeader_p)++;
	}

	//return
	EXIT_CRITICAL();
    return SUCCESS;

end_failCondition:
	//return
	EXIT_CRITICAL();
    return FAIL;

end_breakDownCondition:
	queueReset(queueHeader_p);

	EXIT_CRITICAL();
    return FAIL;

}
/**************************************************************************
* Function     : queueRemove()
*
* Description  : remove the packet from the queue.
* Input        : *packetHeader_p    : pointer to the packet to remove
* Output  	   : SUCCESS/FAIL
* Note         : Its the user responsibility to free the memory allocated in
*                the packet it self, and set the packet status to empty
*
* Note         : This function will remove the active packet.
*
* Revision History:
*		khaled Saab     Oct 29th 2010   initial version
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(queueRemove,"AdaptiveCommonCode")
#endif
uint16_t queueRemove(PACKET_HEADER *packet_p) {

	uint16_t i;
	uint16_t nbPacketsInQueue;	//tmp place holder for queueLength

    QUEUE_HEADER  *queueHeader_p;
	PACKET_HEADER *currPacket_p;
    PACKET_HEADER *nextPacket_p;

	VAR_CRITICAL;		//save and restore HWI info

	//sanity check
	//if the queue empty return fail
	//if the packet is not in the queue, return fail
	//Note: in case of error, we could re-initalize the
	//		entire queue by calling queueInitToDefault()
	if(!packetValid(packet_p))							return FAIL;

	//access to shared resources needs to be performed atomically
	ENTER_CRITICAL();

	queueHeader_p = packetQueue(packet_p);
	currPacket_p  = queueNext(queueHeader_p);

	if(!queueValid (queueHeader_p)) 					goto end_failCondition;
	if( queueLength(queueHeader_p) == 0)				goto end_failCondition;
    if(!packetValid(currPacket_p))						goto end_failCondition;
	if( packetQueue(currPacket_p) != queueHeader_p)		goto end_breakDownCondition;

	//if its the first item in the queue
	if(currPacket_p == packet_p) {
		queueNext(queueHeader_p) = packetNext(packet_p);
		packetNext(packet_p)     = NULL;
		packetQueue(packet_p)    = NULL;//packet does not belong to the
										//queue anymore
		queueLength(queueHeader_p) --;
	}

	//its not the first item in the list
	//search for the packet in the queue
	else {
		currPacket_p     = queueNext(queueHeader_p);//reset to top
		nextPacket_p     = queueNext(queueHeader_p);//reset to top
		nbPacketsInQueue = queueLength(queueHeader_p);
	    for(i = 0; i < nbPacketsInQueue; i++) {

			currPacket_p = nextPacket_p;			//packet IS valid
			nextPacket_p = packetNext(currPacket_p);

			if(!packetValid(nextPacket_p))	//reached end of queue and
				goto end_failCondition;		//the packet is not found
			if(packetQueue(nextPacket_p) != queueHeader_p)
				goto end_breakDownCondition;//sanity check
			if(nextPacket_p == packet_p)	//found the packet
				break;
		}//search for packet

		//the packet is in the queue
		//  remove it
		if(nextPacket_p == packet_p) {
			packetNext(currPacket_p) = packetNext(nextPacket_p);
			packetNext(packet_p)     = NULL;
			packetQueue(packet_p)    = NULL;//packet does not belong to the
											//queue anymore
			queueLength(queueHeader_p) --;
		}
	}//not corner cases

	//if packet is found and its the active packet
	//  disable it
	if(queueActivePacket(queueHeader_p) == packet_p)
		queueActivePacket(queueHeader_p) = NULL;

	//return
	EXIT_CRITICAL();
    return SUCCESS;

end_failCondition:
	EXIT_CRITICAL();
    return FAIL;

end_breakDownCondition:
	queueReset(queueHeader_p);

	EXIT_CRITICAL();
    return FAIL;

}
/**************************************************************************
* Function     : queueMove()
*
* Description  : move a packet from 1 queue to another
* Input        : *packetHeader_p    : pointer to the packet to move
*                *queueHeaderDst_p  : pointer to the destination queue header
* Output  	   : SUCCESS/FAIL
* Note         : The sequence should NOT be changed:
*					1) remove the packet from the source queue
*					2) add the packet to the destination queue
*
* Revision History:
*	Khaled Saab     Oct 29th 2010   initial version
*	Khaled Saab		June 16th 2011	called queueIsMember( ) to make sure that
*									the packet is in the queue before we
*									start moving it around.
*	Khaled Saab		March 8th 2012 	Replace the queueIsMember by a direct
*									call to packetQueue( ) (see changes
*									to the packet header from Jan 11th
*									2012).
**************************************************************************/
#ifdef __TI_COMPILER_VERSION__
#pragma CODE_SECTION(queueMove,"AdaptiveCommonCode")
#endif
uint16_t queueMove( PACKET_HEADER *packet_p,
					QUEUE_HEADER *queueHeaderDst_p) {

	uint16_t status;
	QUEUE_HEADER *queueHeaderSrc_p;

	//sanity check
	//make sure the packet is in the src queue
	if(!packetValid(packet_p))						return FAIL;

	queueHeaderSrc_p = packetQueue(packet_p);
	if(!queueValid(queueHeaderSrc_p)) 				return FAIL;
	if(!queueValid(queueHeaderDst_p)) 				return FAIL;

	//check if there is enough space in the destination queue
	//before moving it
	if(queueLength(queueHeaderDst_p) >= queueMaxLength(queueHeaderDst_p))
		return FAIL;

	//remove the packet from the src queue.
	//if we fail to remove it from the queue
	//	return fail
	//Note: queueRemove function will not delete the packet itself.
	if(queueRemove(packet_p) == FAIL) {
		status = FAIL;
	}

	//add the packet into its destination queue
	//if the add to the destination queue fails
	//	try to added it back to the src queue
	//  return fail
	else if(queueAdd(queueHeaderDst_p, packet_p) == FAIL) {
		status = queueAdd(queueHeaderSrc_p, packet_p);
		if(status == FAIL)
		status = FAIL;
	}
	else
		status = SUCCESS;

	return status;
}
