/*----------------------------------------------------------------------------
 *      RL-ARM - RTX
 *----------------------------------------------------------------------------
 *      Name:    RTX_CONFIG.C
 *      Purpose: Configuration of RTX Kernel for Cortex-M
 *      Rev.:    V4.05
 *----------------------------------------------------------------------------
 *      This code is part of the RealView Run-Time Library.
 *      Copyright (c) 2004-2009 KEIL - An ARM Company. All rights reserved.
 *---------------------------------------------------------------------------*/
#include <RTL.h>

#ifdef CORTEX_APP
#include "../main/main.h"
#include "../../Shared/queueAPI.h"
#endif

/*----------------------------------------------------------------------------
 *      RTX User configuration part BEGIN
 *---------------------------------------------------------------------------*/

//-------- <<< Use Configuration Wizard in Context Menu >>> -----------------
//
// <h>Task Configuration
// =====================
//
//   <o>Number of concurrent running tasks <0-250>
//   <i> Define max. number of tasks that will run at the same time.
//   <i> Default: 6
//Note: if we increase the number of tasks, we need to make sure that
//		there is enough stack to hold them all
//		mp_stk (in .bss) =  OS_STKSIZE *4 * (OS_TASKCNT - OS_PRIVCNT + 1)
#ifndef OS_TASKCNT
 #define OS_TASKCNT     24
#endif

//   <o>Number of tasks with user-provided stack <0-250>
//   <i> Define the number of tasks that will use a bigger stack.
//   <i> The memory space for the stack is provided by the user.
//   <i> Default: 0
#ifndef OS_PRIVCNT
 #define OS_PRIVCNT     0
#endif

//   <o>Task stack size [bytes] <20-4096:8><#/4>
//   <i> Set the stack size for tasks which is assigned by the system.
//   <i> Default: 200
//Note: if we increase the task stack size, we need to make sure that
//		there is enough stack to hold them all
//		mp_stk (in .bss) =  OS_STKSIZE *4 * (OS_TASKCNT - OS_PRIVCNT + 1)
#ifndef OS_STKSIZE
 #define OS_STKSIZE     200
#endif

// <q>Check for the stack overflow
// ===============================
// <i> Include the stack checking code for a stack overflow.
// <i> Note that additional code reduces the Kernel performance.
#ifndef OS_STKCHECK
 #define OS_STKCHECK    1
#endif

// <q>Run in privileged mode
// =========================
//In privileged mode user may access and configure the system and
//control registers like NVIC interrupt controller etc. This is
//however not allowed from unprivileged mode. An access to NVIC
//registers from unprivileged mode will result in Hard Fault.
// <i> Run all Tasks in privileged mode.
// <i> Default: Unprivileged (all tasks to run in protected mode.
//							  The tasks are not allowed to change
//							  system settings, change interrupts etc.)
#ifndef OS_RUNPRIV
 #define OS_RUNPRIV     0
#endif

// </h>
// <h>SysTick Timer Configuration
// =============================
//   <o>Hardware timer <0=> Core SysTick <1=> Peripheral Timer
//   <i> Define the on-chip timer used as a time-base for RTX.
//   <i> Default: Core SysTick
#ifndef OS_TIMER
 #define OS_TIMER       0
#endif

//The timer clock value specifies the input clock frequency and
//depends on CPU clock and APB clock. For a device with CPU clock
//60 MHz and VPB divider 4 the peripheral clock is 15MHz and
//therefore the value is set to 15000000.
//   <o>Timer clock value [Hz] <1-1000000000>
//   <i> Set the timer clock value for selected timer.
//   <i> Default: 16000000  (16MHz)
#ifndef OS_CLOCK
 #define OS_CLOCK       25000000
#endif

//The time tick value specifies the interval of the periodic
//RTX interrupt. The value 10000 us configures timer tick
//period of 0.01 seconds.
//   <o>Timer tick value [us] <1-1000000>
//   <i> Set the timer tick value for selected timer.
//   <i> Default: 10000  (10ms)
#ifndef OS_TICK
 #define OS_TICK        10000
#endif

// </h>

// <h>System Configuration
// =======================
// <e>Round-Robin Task switching
// =============================
// <i> Enable Round-Robin Task switching.
#ifndef OS_ROBIN
 #define OS_ROBIN       1
#endif

//   <o>Round-Robin Timeout [ticks] <1-1000>
//   <i> Define how long a task will execute before a task switch.
//   <i> Default: 5
#ifndef OS_ROBINTOUT
 #define OS_ROBINTOUT   5
#endif

// </e>

//   <o>Number of user timers <0-250>
//   <i> Define max. number of user timers that will run at the same time.
//   <i> Default: 0  (User timers disabled)
#ifndef OS_TIMERCNT
 #define OS_TIMERCNT    4
#endif

//The isr_ library function, when called from the interrupt handler,
//stores the request type and optional parameter to the ISR FIFO Queue
//buffer to be processed later, after the interrupt handler exits.
//OS_FIFOSZ specifies the number of entries that can be stored to
//the FIFO Queue buffer.
//   <o>ISR FIFO Queue size<4=>   4 entries  <8=>   8 entries
//                         <12=> 12 entries  <16=> 16 entries
//                         <24=> 24 entries  <32=> 32 entries
//                         <48=> 48 entries  <64=> 64 entries
//                         <96=> 96 entries
//   <i> ISR functions store requests to this buffer,
//   <i> when they are called from the iterrupt handler.
//   <i> Default: 16 entries
#ifndef OS_FIFOSZ
 #define OS_FIFOSZ      96
#endif

// </h>

//------------- <<< end of configuration section >>> -----------------------

// Standard library system mutexes
// ===============================
//  Define max. number system mutexes that are used to protect
//  the arm standard runtime library. For microlib they are not used.
#ifndef OS_MUTEXCNT
 #define OS_MUTEXCNT    8
#endif

/*----------------------------------------------------------------------------
 *      RTX User configuration part END
 *---------------------------------------------------------------------------*/

#define OS_TRV          ((U32)(((double)OS_CLOCK*(double)OS_TICK)/1E6)-1)

/*----------------------------------------------------------------------------
 *      Global Functions
 *---------------------------------------------------------------------------*/

/*--------------------------- os_idle_demon ---------------------------------*/
__task void os_idle_demon (void) {

	/* The idle demon is a system task, running when no other task is ready */
	/* to run. The 'os_xxx' function calls are not allowed from this task.  */

  for (;;) {
	/* HERE: include optional user code to be executed when no task runs.*/

  }//forever loop
}

/*--------------------------- os_tick_init ----------------------------------*/
#if (OS_TIMER != 0)
int os_tick_init (void) {
  /* Initialize hardware timer as system tick timer. */
  /* ... */
  return (-1);  /* Return IRQ number of timer (0..239) */
}
#endif

/*--------------------------- os_tick_irqack --------------------------------*/
#if (OS_TIMER != 0)
void os_tick_irqack (void) {
  /* Acknowledge timer interrupt. */
  /* ... */
}
#endif

/*--------------------------- os_tmr_call -----------------------------------*/
void os_tmr_call (U16 info) {
  /* This function is called when the user timer has expired. Parameter   */
  /* 'info' holds the value, defined when the timer was created.          */

  /* HERE: include optional user code to be executed on timeout. */
}


/*--------------------------- os_error --------------------------------------*/

void os_error (U32 err_code) {
//This function is called when a runtime error is detected. Parameter
//'err_code' holds the runtime error code (defined in RTL.H).

//HERE: include optional code to be executed on runtime error.

	switch (err_code) {
	case OS_ERR_STK_OVF:
	break;
	case OS_ERR_FIFO_OVF:
	break;
	case OS_ERR_MBX_OVF:
	break;
	default:
	break;
	}//switch


//	 __breakpoint(0); 		//XXX FOR DEBUG: TO BE REMOVED
}


/*----------------------------------------------------------------------------
 *      RTX Configuration Functions
 *---------------------------------------------------------------------------*/

#include <RTX_lib.c>

/*----------------------------------------------------------------------------
 * end of file
 *---------------------------------------------------------------------------*/
