/****************************************************************
  *
  * Proprietary Data of Adaptive Networks, Inc.
  *              All rights reserved
  *
  * Notice:
  *        This file contains trade secret data which
  *        is the property of Adaptive Networks, Inc.  This
  *        file may not be used for any purpose unless
  *        permission is granted in writing by an officer
  *        of Adaptive Networks, Inc.
  *
  *
  * Filename
  *        Ethernet_Task.c
  *
  * Description: Ethernet task manager C code
  *
  * Revision History:
  * Khaled Saab Mar 15th 2016 	Split from taskManager.c
  ****************************************************************/
#include <stdio.h>
#include <stdbool.h>
#include <stdint.h>             //definitions like uint32_t, ...
#include <stdlib.h>				//rand
#include <RTL.h>				//mutex
#include <string.h>				//memcpy

#include <inc/hw_memmap.h>
#include <inc/hw_types.h>
#include <inc/hw_uart.h>
#include <inc/hw_gpio.h>        //GPIO base addresses
#include <inc/hw_ssi.h>
#include <inc/hw_emac.h>
#include <driverlib/emac.h>
#include <driverlib/interrupt.h>
#include <driverlib/uart.h>
#include <driverlib/gpio.h>     //GPIO definitions
#include <driverlib/udma.h>
#include <driverlib/sysctl.h>

#include "../../Shared/DSP1000.h"
#include "../../Shared/interfaces.h"
#include "../../Shared/MacLink.h"
#include "../../Shared/packetAPI.h"
#include "../../Shared/queueAPI.h"

#include "../main/main.h"			//stats
#include "../our_Drivers/taskMgr.h"
#include "../our_Drivers/Ethernet_Driver.h"


/**************************************************************************
* Function     : taskEthernetRx()
*
* Description  : Ethernet Rx Task.
*                This task is the right hand of Ethernet_IRQHandler( ).
*                Ethernet_IRQHandler sets the flags, and the Task do
*                the processing.
*
*       taskEthernetRx( )
*			if its a new packet
*				if (Rx ISR disabled & there is a packet in Rx FIFO)
*					- add the packet to the Ethernet Rx queue
*					- set the active packet
*					- set uDMA to read the packet from Rx FIFO
*			if its the end of a previous receive
*				- set the packet as ready
*				- clear the active packet
*           process all READY packets in the Ethernet Rx queue
*
* Input        :
* Output  	   :
* Revision History:
*		khaled Saab     Nov 18th 2010   creation date
*		Khaled Saab		April 9th 2011 	added checks for error and
*										correspoding recovery code.
**************************************************************************/
__task void taskEthernetRx(void) {

	uint8_t			done;
	uint32_t		status;
	 int32_t		i32FrameLen;
	PACKET_HEADER  *packet_p;
	
    for( ; ; ) {

        //wait on the Rx event
        os_evt_wait_or(TASK_EVENT_RX_FLAG, TASK_ETHERNET_RX_TIMEOUT);

		// Make sure that we own the receive descriptor.
		if((g_psRxDescriptor[g_ui32RxDescIndex].ui32CtrlStatus & DES0_RX_CTRL_OWN))
			goto taskEthernetRxEnd;

		// We own the receive descriptor so check to see if it contains a valid
		// frame.  Look for a descriptor error, indicating that the incoming
		// packet was truncated or, if this is the last frame in a packet,
		// the receive error bit.
		if((g_psRxDescriptor[g_ui32RxDescIndex].ui32CtrlStatus & DES0_RX_STAT_ERR))
			goto taskEthernetRxNext;

		// We have a valid frame. First check that the "last descriptor"
		//flag is set. We sized the receive buffer such that it can always
		//hold a valid frame so this flag should never be clear at this point
		//but...
		if(!g_psRxDescriptor[g_ui32RxDescIndex].ui32CtrlStatus & DES0_RX_STAT_LAST_DESC)
			goto taskEthernetRxNext;

		// Copy the content to the supplied buffer.
		i32FrameLen = ((g_psRxDescriptor[g_ui32RxDescIndex].ui32CtrlStatus &
						DES0_RX_STAT_FRAME_LENGTH_M) >> DES0_RX_STAT_FRAME_LENGTH_S);
		if(i32FrameLen == 0)
			goto taskEthernetRxNext;

		// allocate a packet for the new data
		packet_p = packetAcquire(i32FrameLen);
		status = queueAdd(&gQueueEthernetRx, packet_p);

		//if we don't have enough memory for the packet
		//	reschedule the task and try again
		if(status == FAIL) {
			packetRemove(packet_p);
			goto taskEthernetRxEnd;;
		}
		
		//copy the Ethernet frame from the global buffer to the packet
		memcpy(packetGetPtr(packet_p), g_pui8RxBuffer, i32FrameLen);

		packetLength(packet_p)  = i32FrameLen;
		packetDstEP (packet_p)  = ENDPOINT_POWERLINE;
		packetSrcEP (packet_p)  = ENDPOINT_ETHERNET;
		packetCmd   (packet_p)  = CMD_SEND_PACKET;
		packetState (packet_p) |= PACKET_READY_STATE;

        //process all READY packets in the Rx queue
        done = FALSE;
        while(!done) {
            packet_p = queueGet(&gQueueEthernetRx,
								PACKET_IN_USE_STATE | PACKET_READY_STATE,
								NULL);
            if(packet_p == NULL) done = TRUE;
            else                 EthernetPacketProcess(packet_p);
        }//while we have ready packets in the queue


		taskEthernetRxNext:
        // Move on to the next descriptor in the chain.
        g_ui32RxDescIndex++;
        if(g_ui32RxDescIndex == NUM_RX_DESCRIPTORS)
            g_ui32RxDescIndex = 0;

        // Mark the next descriptor in the ring as available for the receiver
        // to write into.
        g_psRxDescriptor[g_ui32RxDescIndex].ui32CtrlStatus = DES0_RX_CTRL_OWN;

		taskEthernetRxEnd:
		continue;
	}//forever loop
}
/**************************************************************************
* Function     : taskEthernetTx()
*
* Description  : send a packet to the Ethernet Tx FIFO using uDMA
*                   if Tx uDMA is busy
*						reschedule the task
*					if there is no active packet
*						get a new packet from Tx Queue
*						set the active packet
*						send it using uDMA
*					if there is an active packet
*						remove the packet from the queue
*						clear the active packet
* Input        :
* Output  	   :
* Revision History:
*	Khaled Saab     Nov 18th 2010   creation date
*	Khaled Saab		Apr 09th 2011 	added checks for error and
*									correspoding recovery code.
*	Khaled Saab		Jan 17th 2014	moved the code that activates the
*									Ethernet transmit from the ISR to
*									here (MAC_TR_NEWTX)
**************************************************************************/
__task void taskEthernetTx(void) {

 	PACKET_HEADER *packet_p;
	
    for( ; ; ) {

         //wait on the Tx event
        os_evt_wait_or(TASK_EVENT_TX_FLAG, TASK_ETHERNET_TX_TIMEOUT);

		// Wait for the previous packet to be transmitted.
		if(g_psTxDescriptor[g_ui32TxDescIndex].ui32CtrlStatus & DES0_TX_CTRL_OWN)
		{
			continue;
		}

		// Get the packet to transmit
        packet_p = queueGet(&gQueueEthernetTx,
							PACKET_IN_USE_STATE | PACKET_READY_STATE,
							NULL);
        if(packet_p == NULL) continue;
		queueActivePacket(&gQueueEthernetTx) = packet_p;

		// Check that we're not going to overflow the transmit buffer.  This
		// shouldn't be necessary since the uIP buffer is smaller than our DMA
		// transmit buffer but, just in case...
		if(packetLength(packet_p) > TX_BUFFER_SIZE)
		{
			//remove the packet (we are done with it)
			queueRemove(packet_p);
			packetRemove(packet_p);
		}

		// Copy the Ethernet packet to the transmit buffer
		memcpy(g_pui8TxBuffer, packetGetPtr(packet_p), packetLength(packet_p));

		// Move to the next descriptor.
		g_ui32TxDescIndex++;
		if(g_ui32TxDescIndex == NUM_TX_DESCRIPTORS)
			g_ui32TxDescIndex = 0;

		// Fill in the packet size and tell the transmitter to start work.
		g_psTxDescriptor[g_ui32TxDescIndex].ui32Count = packetLength(packet_p);
		g_psTxDescriptor[g_ui32TxDescIndex].ui32CtrlStatus =
								(	DES0_TX_CTRL_LAST_SEG      |
									DES0_TX_CTRL_FIRST_SEG     |
									DES0_TX_CTRL_INTERRUPT     |
//									DES0_TX_CTRL_IP_ALL_CKHSUMS|
									DES0_TX_CTRL_NO_CHKSUM     |
									DES0_TX_CTRL_CHAINED       |
									DES0_TX_CTRL_OWN);

		// Tell the DMA to reacquire the descriptor now that we've filled it in.
		EMACTxDMAPollDemand(EMAC0_BASE);

		//remove the packet (we are done with it)
		queueRemove(packet_p);
		packetRemove(packet_p);
	}//forever loop
}
