/****************************************************************
 *
 *    Proprietary Data of Adaptive Networks Inc.
 *              All rights reserved
 *
 * Notice:
 *		This file contains trade secret data which
 *		is the property of Adaptive Networks Inc.  This
 *		file may not be used for any purpose unless
 *		permission is granted in writing by an officer
 *		of Adaptive Networks Inc.
 *
 *
 * Filename
 *		Ethernet_Udp.c
 *
 * Description: udp function
 *					- UDPPacketAcquire(): create a udp packet
 *
 * History:
 *		Siyi Tan	Jan 10th 2013   original version
 ****************************************************************/
#ifndef __TI_COMPILER_VERSION__
#include <rtl.h>
#endif

#include <string.h>				//definitions memcpy

#include "../StellarisWare/lwip-1.3.2/src/include/ipv4/lwip/inet_chksum.h"
#include "../our_Drivers/Ethernet_Driver.h"
#include "../our_Drivers/Ethernet_Udp.h"
#include "../../Shared/DSP1000.h"
#include "../../Shared/packetAPI.h"


 /**************************************************************************
 * Function     : UDPPacketAcquire
 * Description  : Generate an ethernet packet
 * Input        : data_p     : payload of the udp packet
 *				  dataLength : the length of the payload
 *				  destipaddr : the destination IP address
 *				  srcipaddr  : the source IP address
 *				  destMacAddr: the destination MAC address
 *				  destUdpPort: the destination UDP port
 *				  srcUdpPort : the source UDP port
 *				  ipId       : IP ID			 
 *				  srcEp      : source endpoint
 *				  dstEp 	 : destination endpoint
 * 				  cmd        : packet command [	CMD_SNMP_PACKET, 
 *												CMD_SEND_MGMT...]
 * Output       : packet_p   : created packet
 * Return       : packet_p   : created packet
 * History      : 
 * Khaled Saab  Oct 19th 2012   Creation date.
 * Siyi Tan	  	Jan 03th 2012	Modifed.
 * Khaled Saab	Mar  2th 2016	Added input arguments srcEp, dstEp,
 *								and cmd for extra flexibility.
**************************************************************************/
PACKET_HEADER *UDPPacketAcquire(uint8_t *data_p, 
						 uint16_t dataLength, 
						 uint8_t *destipaddr,
						 uint8_t *srcipaddr,
						 uint8_t *destMacAddr,
						 uint8_t *srcMacAddr,
						 uint8_t *destUdpPort,
						 uint8_t *srcUdpPort,
						 uint8_t *ipId,
						 uint8_t  srcEp,
						 uint8_t  dstEp,
						 uint16_t cmd) { 


	struct ethip_hdr    *ipHeader_p;	//ethernet Ip header
	struct udp_hdr		*udpHeader_p;	//ethernet udp header
	PACKET_HEADER 		*packet_p;
	uint8_t 			*dataLocal_p;	//data pointer in the packet
	uint16_t 			packetLength;

	packetLength = dataLength + sizeof(struct ethip_hdr) + sizeof(struct udp_hdr) ;
	
	//allocate memory for hte packet
	packet_p = packetAcquire(packetLength);

	if(packet_p == NULL){
		return NULL;
	}

	//fill the data
	dataLocal_p = packetPtr(packet_p, sizeof(struct ethip_hdr) + sizeof(struct udp_hdr));
	memcpy(dataLocal_p, data_p, dataLength);
	
	//fill the udp header and the ip header
	ipHeader_p  = (struct ethip_hdr *) packetGetPtr(packet_p);
	udpHeader_p = (struct udp_hdr   *) packetPtr(packet_p, sizeof(struct ethip_hdr));
	
	udpHeader_p->srcport   = *((uint16_t*)srcUdpPort);	 
	udpHeader_p->destport  = *((uint16_t*)destUdpPort);
	udpHeader_p->udplen    = htons(sizeof(struct udp_hdr) + dataLength);
	udpHeader_p->udpchksum = 0;
	udpHeader_p->udpchksum = inet_chksum_pseudo((void*)udpHeader_p,
       											srcipaddr, 
												destipaddr,
       											UIP_PROTO_UDP,
												sizeof(struct udp_hdr) + dataLength);
														

	memcpy(ipHeader_p->ethhdr.src.addr , srcMacAddr , 6);
	memcpy(ipHeader_p->ethhdr.dest.addr, destMacAddr, 6);
    memcpy(ipHeader_p->srcipaddr       , srcipaddr	, 4);
	memcpy(ipHeader_p->destipaddr      , destipaddr	, 4);

	ipHeader_p->ethhdr.type = htons(UIP_ETHTYPE_IP);
	ipHeader_p->vhl         = 4 << 4 ;		//ipv4
	ipHeader_p->vhl        |= ((sizeof(struct ethip_hdr) - sizeof(struct uip_eth_hdr)) >> 2);	//should be 5
	ipHeader_p->tos 		= 0;
	ipHeader_p->len[0]      = (sizeof(struct ip_hdr) + sizeof(struct udp_hdr) + dataLength) >> 8;
	ipHeader_p->len[1]      = (sizeof(struct ip_hdr) + sizeof(struct udp_hdr) + dataLength) & 0xFF;

	if (ipId!= NULL){
		ipHeader_p->ipid[0]     = ipId[0];	
		ipHeader_p->ipid[1]     = ipId[1];	
	}

	ipHeader_p->ipoffset[0] = 0;
	ipHeader_p->ipoffset[1] = 0;
	ipHeader_p->ttl         = IP_DEFAULT_TTL;
	ipHeader_p->proto       = UIP_PROTO_UDP;
	ipHeader_p->ipchksum	= 0;
	ipHeader_p->ipchksum    = inet_chksum((void*)((uint32_t)ipHeader_p + sizeof(struct uip_eth_hdr)),
										   sizeof(struct ip_hdr));
	
	//complete the packet setup
	packetLength(packet_p)  = sizeof(struct ethip_hdr) + sizeof(struct udp_hdr) + dataLength;
	packetCmd(   packet_p)  = cmd;
	packetSrcEP( packet_p)  = srcEp;
	packetDstEP( packet_p)  = dstEp;
	packetState( packet_p) |= PACKET_READY_STATE;
	
	//return
	return packet_p;
	}
